;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((nil
  (fill-column . 80)
  (c-file-style . "bsd")
  (c-basic-offset . 4)
  (tab-width . 4))
 (cmake-mode
  (cmake-tab-width . 4)))
